@@ -8,6 +8,7 @@ NOTE: Using this tool requires setting the GOOGLE_API_KEY and GOOGLE_CSE_ID
 environment variables in your `.env` file, as explained in the
 [README](https://github.com/langroid/langroid#gear-installation-and-setup).
 """
+from typing import List
 
 from langroid.agent.tool_message import ToolMessage
 from langroid.parsing.web_search import google_search
@@ -26,3 +27,12 @@ class GoogleSearchTool(ToolMessage):
         search_results = google_search(self.query, self.num_results)
         # return Title, Link, Summary of each result, separated by two newlines
         return "\n\n".join(str(result) for result in search_results)
+
+    @classmethod
+    def examples(cls) -> List["ToolMessage"]:
+        return [
+            cls(
+                query="When was the Llama2 Large Language Model (LLM) released?",
+                num_results=3,
+            ),
+        ]
